#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <packet.h>
#include <preproc.h>
#include <capdev.h>
#include <decode.h>
#include <alert.h>
#include <plugin.h>

/* 
 * IP Defragmentation for firestorm
 * ================================
 * 
 * This code is a reworking of ip_fragment.c from Linux 2.4.18,
 * it should be relatively straight forward to understand. This
 * is a first cut so it still might not be perfect.
 *
 * Should cope with:
 *  Overlapping fragments
 *  Oversized fragments
 *  Out of order fragments
 *
 * TODO:
 *  evictor optimisations (global lru/cache tail)
 *  Timeout packets (worth bothering?)
 *  Coalesce fragments in incomplete packets (??)
 */

#if 1
#define dprintf(x, ...) 
#define __INLINE__ static inline
#else
#define dprintf printf
#define __INLINE__
#endif

/* Statistics */
unsigned int ipfrag_errs=0;
unsigned int ipfrag_reassembled=0;

/* Don't decode fragments inside fragments */
serial_t last_frag;
unsigned int first_frag=0;

/* Memory usage, high and low water marks */
unsigned int ipfrag_mem=0;
unsigned int ipfrag_mem_hi=1024*1024;
unsigned int ipfrag_mem_lo=768*1024;

struct {
	char *name;
	unsigned int *val;
}ipfrag_args[]={
	{"mem_hi", &ipfrag_mem_hi},
	{"mem_lo", &ipfrag_mem_lo},
	{NULL, NULL}
};

/* Stuff we import */
proc_preproc_activate preproc_activate;
proc_preproc_dispatch preproc_dispatch;
proc_decode_proto decode_proto;
proc_serial_number serial_number;
proc_alert alert;
struct proto *ipproto;

/* Mask to determine fragments */
unsigned short ipfmask=__constant_htons(IP_MF|IP_OFFMASK);

/* Reassembly */
struct packet ipfr;
char *reasm_buf=NULL;
unsigned int reasm_len=0;

/* Keeps each individual fragment */
struct ipfrag {
	struct ipfrag		*next;
	unsigned int		len;
	unsigned int		offset;
	void			*data;
	unsigned int		free;
	void			*fdata; /* Data to free */
};

struct ipq {
	struct ipq	*next;
	struct ipq	**pprev;

	/* Identify the packet */
	u_int32_t	saddr;
	u_int32_t	daddr;
	u_int16_t	id;
	u_int8_t	protocol;

	/* Linked list of fragments */
	struct ipfrag	*fragments;
	unsigned int	num_frags;

#define FIRST_IN	2
#define LAST_IN		1
	int		last_in;

	/* Total size of all the fragments we have */
	int		meat;

	/* Total length of full packet */
	int		len;

	/* Stuff we need for reassembly */
	struct cap_s	*cap;
	u_int32_t	flags;
	struct timeval	time;
};

#define IPHASH 128 /* Must be a power of two */
struct ipq *ipq_hash[IPHASH]; /* IP fragment hash table */

/* Hash function for ipq_hash lookup */
__INLINE__ unsigned int ipq_hashfn(u_int16_t id,
	u_int32_t saddr,
	u_int32_t daddr,
	u_int8_t proto){
	unsigned int h=saddr^daddr;

	h ^= (h>>16)^id;
	h ^= (h>>8)^proto;
	return h & (IPHASH-1);
}

/*
 * Report ip fragmentation violations.
 */
void ipfrag_teardrop(struct packet *p)
{
	static struct alert ta={alert:"Teardrop"};
	alert("ipfrag", p, &ta);
}
void ipfrag_oversize(struct packet *p)
{
	static struct alert ta={alert:"Oversized fragments"};
	alert("ipfrag", p, &ta);
}
void ipfrag_attack(struct packet *p)
{
	static struct alert ta={alert:"Fragmentation attack"};
	alert("ipfrag", p, &ta);
}
void ipfrag_boink(struct packet *p)
{
	static struct alert ta={alert:"Boink"};
	alert("ipfrag", p, &ta);
}

void ipq_kill(struct ipq *qp)
{
	struct ipfrag *foo, *bar;

	/* Unlink from the list */
	if ( qp->next )
		qp->next->pprev=qp->pprev;
	*qp->pprev=qp->next;

	/* Free the fragments and descriptors */
	for(foo=qp->fragments; foo;) {
		bar=foo;
		foo=foo->next;
		if ( bar->free ) free(bar->fdata);
		free(bar);
		ipfrag_mem-=sizeof(struct ipfrag);
	}

	/* Free the ipq itself */
	free(qp);
	ipfrag_mem-=sizeof(struct ipq);
}

/* Create a checksum for the ip header. I don't know
 * why we bother really... */
void ipfrag_csum(struct pkt_iphdr *iph)
{
	u_int16_t *tmp=(u_int16_t *)iph;
	u_int32_t sum=0;
	int i;

	for(i=0; i<iph->ihl<<1; i++) {
		/* ignore field 5, its the checksum ;) */
		if ( i!=5 ) sum+=tmp[i];
		if(sum & 0x80000000) {
			sum = (sum & 0xffff) + (sum >> 16);
		}
	}

	while(sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);
	
	iph->csum=(~sum) & 0xffff;
}

/* Reassemble a complete set of fragments, decode the
 * new packet, and send it back through the preprocessor
 * list, we won't touch it next time round */
__INLINE__ void ipfrag_reassemble(struct ipq *qp)
{
	struct ipfrag *f;
	struct pkt_iphdr *iph;
	unsigned int olen=qp->len;
	unsigned int len=0;
	unsigned int hlen=0;
	char *buf;

	/* Kill oversize packets */
	if ( olen > 65535 ) {
		ipfrag_errs++;
		ipq_kill(qp);
		return;
	}

	iph=qp->fragments->data;
	hlen=iph->ihl<<2;
	qp->fragments->len+=hlen;
	olen+=hlen;

	/* Allocate the frankenpacket buffer */
	if ( reasm_buf ) {
		char *tmp;
		
		/* Expand the buffer if it's too small */
		if ( olen < reasm_len ) {
			if ( !(tmp=realloc(reasm_buf, olen)) ) {
				olen=reasm_len;
			}else{
				reasm_buf=tmp;
				reasm_len=qp->len;
			}
		}
	}else{
		if ( !(reasm_buf=malloc(olen)) ) {
			dprintf("ipfrag: reassemble error!\n");
			ipfrag_errs++;
			return;
		}
		reasm_len=olen;
	}
	
	/* Copy all the fragments in to the new buffer */
	dprintf("Reassemble: %u bytes\n", olen);
	buf=reasm_buf;
	for(f=qp->fragments; f; f=f->next) {
		dprintf(" * %u bytes @ %u\n", f->len, f->offset);
		memcpy(buf, f->data, f->len);
		buf+=f->len;
		len+=f->len;
		if ( len >= olen ) break;
	}
	dprintf("\n");

	/* Build the packet */
	serial_number(&ipfr.serial);
	ipfr.len=len;
	ipfr.caplen=len;
	ipfr.llen=1;
	ipfr.layer[0].proto=ipproto;
	ipfr.layer[0].h.raw=reasm_buf;
	ipfr.layer[0].flags=FLAG_IP_REASM|FLAG_IP_CSUM;
	ipfr.base=reasm_buf;
	ipfr.end=ipfr.base+olen;
	ipfr.capture=qp->cap;
	ipfr.flags=qp->flags | FP_CLONE;

	/* Fill in the timestamp */
	ipfr.time.tv_sec=qp->time.tv_sec;
	ipfr.time.tv_usec=qp->time.tv_usec;
	
	/* Kill the queue entry */
	ipq_kill(qp);

	/* Fixup the IP header */
	ipfr.layer[0].h.ip->frag_off=0;
	ipfr.layer[0].h.ip->tot_len=ntohs(len);
	ipfrag_csum(ipfr.layer[0].h.ip);

	/* Inject the packet back in to the flow */
	ipfrag_reassembled++;
	if ( ipfr.layer[0].proto->decode(&ipfr) ) {
		preproc_dispatch(&ipfr);
	}
	
	return;
}

__INLINE__ struct ipq *ip_frag_create(unsigned int hash, struct pkt_iphdr *iph)
{ 
	struct ipq *q;
	
	if ( !(q=calloc(1, sizeof(struct ipq))) ){
		ipfrag_errs++;
		return NULL;
	}
	ipfrag_mem+=sizeof(struct ipq);

	q->id=iph->id;
	q->saddr=iph->saddr;
	q->daddr=iph->daddr;
	q->protocol=iph->protocol;

	if ( (q->next=ipq_hash[hash]) ) {
		q->next->pprev=&q->next;
	}
	ipq_hash[hash]=q;
	q->pprev=&ipq_hash[hash];

	return q;
}

/* Find (or create) the ipq for this IP fragment */
__INLINE__ struct ipq *ip_find(struct pkt_iphdr *iph, unsigned int *hash)
{
	struct ipq *qp;

	*hash=ipq_hashfn(iph->id, iph->saddr,
		iph->daddr, iph->protocol);

	for(qp=ipq_hash[*hash]; qp; qp=qp->next) {
		if ( qp->id == iph->id &&
			qp->saddr == iph->saddr &&
			qp->daddr == iph->daddr &&
			qp->protocol == iph->protocol ) {
			return qp;
		}
	}

	return ip_frag_create(*hash, iph);
}

void ip_evictor(struct ipq *cq)
{
	int i, progress;
	struct ipq *qp;
	
	dprintf("Running the ipfrag evictor!\n");
	do {
		if ( ipfrag_mem <= ipfrag_mem_lo ) return;

		progress=0;

		for(i=0; i<IPHASH; i++) {
			if ( !(qp=ipq_hash[i]) ) continue;

			while(qp->next)
				qp = qp->next;

			if ( qp != cq ) {
				ipfrag_errs++;
				progress=1;
				ipq_kill(qp);
			}
		}
	}while(progress);
}

__INLINE__ void ipfrag_queue(
	unsigned int hash,
	struct ipq *qp, 
	struct packet *pkt, 
	struct pkt_iphdr *iph)
{
	struct ipfrag *prev, *next, *me;
	int flags, offset;
	int ihl, end, len;
	int chop=0;

	/* Remove from queue */
	if ( qp->next )
		qp->next->pprev=qp->pprev;
	*qp->pprev=qp->next;

	/* Prepend to queue */
	if ( (qp->next=ipq_hash[hash]) )
		qp->next->pprev=&qp->next;
	ipq_hash[hash]=qp;
	qp->pprev=&ipq_hash[hash];

	/* Kill off LRU ipqs */
	if ( ipfrag_mem > ipfrag_mem_hi ) {
		ip_evictor(qp);
	}

	offset=ntohs(iph->frag_off);
	flags=offset & ~IP_OFFMASK;
	offset&=IP_OFFMASK;
	offset<<=3; /* 8 byte granularity */

	ihl=iph->ihl<<2;

	len=ntohs(iph->tot_len);

	end = offset + len - ihl;

	if ( (flags&IP_MF)==0 ) {
		if ( end < qp->len ||
			((qp->last_in & LAST_IN) && end!=qp->len)) {
			ipfrag_teardrop(pkt);
			return;
		}
		qp->last_in |= LAST_IN;
		qp->len = end;
	}else{
		if ( end&7 ) {
			/* Don't drop the packet stupid! Modern
			 * stacks mask off 0x7 so if we ditch the
			 * frag as invalid we could be evaded. */
			ipfrag_boink(pkt);
		}

		/* Non-terminal fragments must be multiples of
		 * 8 bytes so mask off low-order bits */
		end &=~7;

		if ( end > qp->len ) {
			if (qp->last_in & LAST_IN) {
				ipfrag_attack(pkt);
				return;
			}
			qp->len=end;
		}
	}

	if ( end==offset ) {
		ipfrag_attack(pkt);
		return;
	}

	/* Don't bother wasting any more resources
	 * when we know the packet is oversize (invalid) */
	if ( qp->len > 65535 ) {
		ipfrag_oversize(pkt);
		return;
	}

	/* Insert data into fragment chain */
	if ( !(me=malloc(sizeof(struct ipfrag))) ) goto ipfrag_oom;
	ipfrag_mem+=sizeof(struct ipfrag);

	/* Find out where to insert this fragment in the list */
	prev=NULL;
	for(next=qp->fragments; next; next=next->next) {
		if ( next->offset >= offset ) break;
		prev=next;
	}

	/* Check we don't overlap the previous fragment */
	if ( prev ) {
		int i=(prev->offset + prev->len) - offset;

		if ( i>0 ) {
			offset += i;

			if ( end <= offset ) {
				ipfrag_attack(pkt);
				return;
			}

			chop=i;
			len-=i;
		}
	}

	/* Make sure we don't overlap next packets */
	while(next && next->offset < end) {
		int i = end - next->offset;

		if ( i < next->len ) {
			/* Eat head of the next overlapped fragment
			 * and leave the loop. The next ones cannot
			 * overlap. */
			next->offset+=i;
			next->len-=i;
			next->data+=i;
			break;
		}else{
			struct ipfrag *free_it=next;
			
			/* Old fragment is completely overriden
			 * with new one. Drop it */
			next=next->next;
			if ( prev ){
				prev->next=next;
			}else{
				qp->fragments=next;
				qp->num_frags--;
			}

			qp->meat -= free_it->len;
			free(free_it);
			ipfrag_mem-=sizeof(struct ipfrag);
		}
	}

	/* Make the fragment */
	me->len=len-(chop+ihl);
	me->offset=offset+chop;

	if ( offset ) chop+=ihl;

	/* IP defragmentation can be zerocopy */
	if ( pkt->flags & FP_CLONE ) {
		int alen=me->len;
		
		if ( !offset ) alen+=ihl;
		
		if ( !(me->fdata=malloc(alen)) ) {
			free(me);
			goto ipfrag_oom;
		}
		memcpy(me->fdata, ((void *)iph)+chop, alen);
		me->free=1; 
	}else {
		me->fdata=((void *)iph)+chop;
		me->free=0;
	}

	me->data=me->fdata;

	/* Insert the fragment */
	me->next=next;
	if ( prev ) {
		prev->next=me;
	}else{
		qp->fragments=me;
		qp->num_frags++;
	}

	/* Finish up */
	qp->cap=pkt->capture;
	qp->flags=pkt->flags;
	qp->meat+=me->len;
	if ( !offset ) qp->last_in |= FIRST_IN;

	/* The time for the reassembled packet is equal
	 * to the time of the last packet recieved. This
	 * makes things sane in the sense that time won't
	 * be seen to be going backwards by the higher layers!
	 */
	qp->time.tv_sec=pkt->time.tv_sec;
	qp->time.tv_usec=pkt->time.tv_usec;

	dprintf("0x%x: got a fragment\n", (unsigned int)qp);

	return;

ipfrag_oom:
	ip_evictor(qp);
	return;
}

/* Preprocessor entry point */
int ipfrag_process(struct packet *pkt, unsigned int i, void *priv)
{
	struct pkt_iphdr *iph;
	unsigned int hash;
	struct ipq *q;

	iph=pkt->layer[i].h.ip;
	if ( iph->frag_off & ipfmask) {

		/* Don't track fragments inside fragments
		 * wait till they are reassembled */
		if ( first_frag && (pkt->serial==last_frag) ) {
			printf("ipfrag: fragment in fragment!\n");
			return PREPROC_OK;
		}

		if ( !(pkt->layer[i].flags&FLAG_IP_CSUM) )
			return PREPROC_OK;

		first_frag=1;
		last_frag=pkt->serial;
		
		if ( (q=ip_find(iph, &hash)) ) {
			ipfrag_queue(hash, q, pkt, iph);
			if ( q->last_in == (FIRST_IN|LAST_IN) &&
				q->meat == q->len ) {
				ipfrag_reassemble(q);
			}
		}
	}

	return PREPROC_OK;
}

/* Parse numbers of the form '3128k', '1234M', '1k4' etc.. */
int ipfrag_val(char *s, unsigned int *v)
{
	unsigned int d;
	int p;

	for(*v=0,p=0; s[p]; p++) {
		if ( (d=(unsigned int)(s[p]-'0')) < 10 ) {
			*v=*v*10+d;
		}else if (s[p]=='k' || s[p]=='K' ) {
			(*v)*=1024;
		}else if (s[p]=='m' || s[p]=='M' ) {
			(*v)*=(1024*1024);
		}else{
			return 0;
		}
	}
	
	return 1;
}

int ipfrag_arg(char *k, char *v)
{
	int i;

	for(i=0; ipfrag_args[i].name; i++) {
		if ( !strcmp(ipfrag_args[i].name, k) ) {
			return ipfrag_val(v, ipfrag_args[i].val);
		}
	}

	return 0;
}

int ipfrag_parseargs(char *args)
{
	int state=0;
	char *cur;
	char *k=NULL, *v=NULL;

	for(cur=args; ;cur++) {
		if ( *cur==0 && state!=2 ) break;

		if ( state==0 ) {
			if ( *cur!=' ' && *cur!='\t' ) {
				k=cur;
				state=1;
			}
		}else if ( state==1 ) {
			if (*cur==' ' || *cur=='\t' ) {
				*cur=0;
			}else if ( *cur=='=' ) {
				*cur=0;
				v=cur+1;
				state=2;
			}
		}else if ( state==2 ) {
			if ( *cur==' ' || *cur=='\t' || *cur==0 ) {
				*cur=0;
				if ( !ipfrag_arg(k, v) ) {
					return 0;
				}
				state=0;
			}
		}
	}
	
	return 1;
}

int ipfrag_init(char *args)
{
	if ( ipproto ) {
		printf("ipfrag: Can't add ipfrag twice!\n");
		return 0;
	}
	
	if ( !(ipproto=decode_proto("ipv4")) ) {
		printf("ipfrag: ipv4 protocol not loaded\n");
		return 0;
	}
	
	if ( !ipfrag_parseargs(args) ) {
		printf("ipfrag: Unable to parse args!\n");
		return 0;
	}

	if ( ipfrag_mem_hi<=ipfrag_mem_lo ) {
		printf("ipfrag: mem_hi must be bigger than mem_lo\n");
		return 0;
	}
	
	if (preproc_activate("ip", ipfrag_process, NULL)) {
		printf("ipfrag: IP defragmentation ENABLED.\n");
		printf("ipfrag: mem_hi=%u mem_lo=%u\n",
			ipfrag_mem_hi, ipfrag_mem_lo);
		return 1;
	}else{
		return 0;
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_preproc_add preproc_add;
	int ok=0;
	
	plugin_check(in, out);
	
	PLUGIN_ID("preproc.ipfrag", "IP defragmentation");
	PLUGIN_VERSION(0, 2);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	if ( !(preproc_add=in->import("preproc.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_activate=in->import("preproc.activate")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_dispatch=in->import("preproc.dispatch")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(decode_proto=in->import("decode.proto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(serial_number=in->import("serial_number")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(alert=in->import("alert")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=preproc_add("ipfrag", ipfrag_init);
	memset(&ipq_hash, 0, sizeof(ipq_hash));
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	if ( ipproto ) {
		printf("ipfrag: %u reassembled packets, "
			"%u errors, %uKB left\n",
			ipfrag_reassembled,
			ipfrag_errs, ipfrag_mem/1024);
	}
	return PLUGIN_ERR_OK;
}
